"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from . import chat_prompt, server
from .chat_prompt import *  # noqa: F403
from .server import *  # noqa: F401, F403

# Combine all exports from submodules
__all__: list[str] = []
__all__.extend(chat_prompt.__all__)
__all__.extend(server.__all__)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from typing import Optional

from a2a.server.agent_execution import AgentExecutor
from a2a.server.tasks import TaskStore
from a2a.types import AgentCard
from a2a.utils.constants import AGENT_CARD_WELL_KNOWN_PATH


@dataclass
class A2APluginOptions:
    agent_card: AgentCard
    "The agent card to be used for the A2A plugin."
    path: Optional[str] = "/a2a"
    "Path to the A2A server, by default is `a2a`."
    agent_card_path: Optional[str] = AGENT_CARD_WELL_KNOWN_PATH
    "Path to the agent card, by default is `/.well-known/agent-card.json`."
    task_store: Optional[TaskStore] = None
    """
    TaskStore which stores the tasks that are sent to the agent or that the agent sends.
    If not provided, the App's storage will be used.
    """
    agent_executor: Optional[AgentExecutor] = None
    """
     For a completely custom executor, you may provide your own executor that will
     get executed whenever the a2a agent is InMemoryTaskStore
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from logging import Logger
from typing import Annotated, Any, Awaitable, Callable, List, Optional

from microsoft_teams.apps import (
    DependencyMetadata,
    EventMetadata,
    HttpPlugin,
    LoggerDependencyOptions,
    Plugin,
    PluginBase,
)

from a2a.server.agent_execution import AgentExecutor
from a2a.server.apps import A2AFastAPIApplication
from a2a.server.request_handlers import DefaultRequestHandler, RequestHandler
from a2a.server.tasks import InMemoryTaskStore, TaskStore
from a2a.types import AgentCard
from a2a.utils.constants import AGENT_CARD_WELL_KNOWN_PATH

from .a2a_plugin_options import A2APluginOptions
from .custom_agent_executor import A2AMessageEvent, CustomAgentExecutor
from .logging_middleware import LoggingMiddleware


@Plugin(name="a2a", version="0.3.7", description="A2A Server Plugin")
class A2APlugin(PluginBase):
    logger: Annotated[Logger, LoggerDependencyOptions()]
    http: Annotated[HttpPlugin, DependencyMetadata()]

    emit: Annotated[Callable[[str, A2AMessageEvent], Awaitable[None]], EventMetadata(name="custom")]

    card: AgentCard
    path: str
    agent_card_path: str
    task_store: TaskStore
    custom_executor: Optional[AgentExecutor] = None
    _middlewares: List[Any] = []

    def __init__(self, options: A2APluginOptions):
        super().__init__()

        self.card = options.agent_card

        if options.path:
            self.path = options.path if options.path.startswith("/") else f"/{options.path}"

        self.agent_card_path = options.agent_card_path if options.agent_card_path else AGENT_CARD_WELL_KNOWN_PATH
        self.task_store = options.task_store if options.task_store else InMemoryTaskStore()
        self.custom_executor = options.agent_executor

    def use(self, middleware: Any, **kwargs: Any) -> None:
        """Add a middleware to the A2A plugin.

        Args:
            middleware: Either a middleware class (inheriting from BaseHTTPMiddleware)
                       or a dispatch function
            **kwargs: Additional keyword arguments to pass to the middleware constructor
        """
        self._middlewares.append((middleware, kwargs))

    async def on_init(self) -> None:
        a2a_app = A2AFastAPIApplication(agent_card=self.card, http_handler=self._setup_request_handler())
        self.app = a2a_app.build(agent_card_url=self.agent_card_path)

        # add the middleware
        self.app.add_middleware(LoggingMiddleware, logger=self.logger)
        for middleware_info in self._middlewares:
            middleware, kwargs = middleware_info
            self.app.add_middleware(middleware, **kwargs)

        self.logger.info(f"A2A agent set up at {self.agent_card_path}")
        self.logger.info(f"A2A agent listening at {self.path}")

        self.http.app.mount(self.path, self.app)

    def _setup_executor(self) -> AgentExecutor:
        return CustomAgentExecutor(self.emit)

    def _setup_request_handler(self) -> RequestHandler:
        return DefaultRequestHandler(
            agent_executor=self.custom_executor if self.custom_executor else self._setup_executor(),
            task_store=self.task_store,
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import re
import uuid
from dataclasses import asdict
from functools import partial
from logging import Logger
from typing import Any, Dict, List, Optional, Union, Unpack, cast

from httpx import AsyncClient
from microsoft_teams.ai import Function as ChatFunction
from microsoft_teams.ai import SystemMessage
from microsoft_teams.ai.plugin import BaseAIPlugin
from microsoft_teams.common.logging.console import ConsoleLogger
from pydantic import BaseModel

from a2a.client import ClientFactory
from a2a.client.card_resolver import A2ACardResolver
from a2a.client.client import ClientConfig
from a2a.types import AgentCard, Message, Part, Role, Task, TextPart

from .agent_client_info import AgentClientInfo
from .agent_config import AgentConfig
from .types import (
    A2AClientPluginOptions,
    A2APluginUseParams,
    AgentPromptParams,
    BuildFunctionMetadata,
    BuildMessageForAgent,
    BuildMessageForAgentMetadata,
    BuildMessageFromAgentMetadata,
    BuildMessageFromAgentResponse,
    BuildPrompt,
    BuildPromptMetadata,
    FunctionMetadata,
    InternalA2AClientPluginOptions,
)


class A2AClientMessageParams(BaseModel):
    message: str
    "The message to send to the agent"


class A2AClientPlugin(BaseAIPlugin):
    log: Logger
    _agent_configs: Dict[str, AgentConfig] = {}
    _clients: Dict[str, AgentClientInfo] = {}

    _build_function_metadata: Optional[BuildFunctionMetadata] = None
    _build_prompt: Optional[BuildPrompt] = None
    _build_message_for_agent: Optional[BuildMessageForAgent] = None
    _build_message_from_agent_response: Optional[BuildMessageFromAgentResponse] = None

    def __init__(self, **options: Unpack[A2AClientPluginOptions]):
        super().__init__("a2a")

        self.options = InternalA2AClientPluginOptions.from_typed_dict(options)
        self._build_function_metadata = self.options.build_function_metadata
        self._build_prompt = self.options.build_prompt
        self._build_message_for_agent = self.options.build_message_for_agent
        self._build_message_from_agent_response = self.options.build_message_from_agent_response
        self.log = self.options.logger if self.options.logger else ConsoleLogger().create_logger(name="a2a:client")

    def on_use_plugin(self, args: A2APluginUseParams):
        # just store the config, defer client creation to on_build_functions
        self._agent_configs.update(
            {
                args.key: AgentConfig(
                    key=args.key,
                    base_url=args.base_url,
                    card_url=args.card_url,
                    build_function_metadata=args.build_function_metadata,
                    build_message_for_agent=args.build_message_for_agent,
                    build_message_from_agent_response=args.build_message_from_agent_response,
                )
            }
        )

    async def fetch_agent_card(self, base_url: str, card_url: str) -> AgentCard:
        async with AsyncClient() as httpx_client:
            resolver = A2ACardResolver(httpx_client, base_url)
            card = await resolver.get_agent_card(card_url)
            return card

    async def _get_agent_card(self, key: str, config: AgentConfig) -> Optional[AgentCard]:
        # return cached client info if it exists
        client_info = self._clients.get(key)

        if client_info:
            return client_info.agent_card

        # create new client and get agent card
        try:
            card = await self.fetch_agent_card(base_url=config.base_url, card_url=config.card_url)
            client_config = ClientConfig()
            client = ClientFactory(client_config).create(card=card)
            client_info = AgentClientInfo(**asdict(config), client=client, agent_card=card)
            self._clients.update({key: client_info})
            return card
        except Exception as e:
            self.log.error(f"Error creating client or fetching agent card for {key}: {e}")
            return None

    def _default_function_metadata(self, card: AgentCard) -> FunctionMetadata:
        name = f"message_{re.sub(r'(?<!^)(?=[A-Z])', '_', card.name).lower()}"
        description = card.description or f"Interact with {card.name} agent"
        return FunctionMetadata(name=name, description=description)

    def _create_message(
        self, text: str, card: Optional[AgentCard] = None, metadata: Optional[Dict[str, Any]] = None
    ) -> Message:
        return Message(
            message_id=str(uuid.uuid4()),
            role=Role("user"),
            parts=[Part(root=TextPart(kind="text", text=text))],
            metadata=metadata,
        )

    def _default_build_message(self, card: AgentCard, input: str, metadata: Optional[Dict[str, Any]] = None) -> Message:
        return self._create_message(input, card, metadata)

    def _default_build_message_from_agent_response(
        self, card: AgentCard, response: Union[Message, Task], original_input: str
    ) -> str:
        if isinstance(response, Message):
            text_parts: List[str] = []
            for part in response.parts:
                if getattr(part.root, "kind", None) == "text":
                    text_part = cast(TextPart, part.root)
                    text_parts.append(text_part.text)
            return " ".join(text_parts) or "Agent responded with no text content."
        else:
            return "Agent responded with non-message content."

    async def on_build_functions(self, functions: list[ChatFunction[Any]]) -> list[ChatFunction[Any]] | None:
        all_functions: List[ChatFunction[Any]] = []

        for key, config in self._agent_configs.items():
            try:
                agent_card = await self._get_agent_card(key, config)
                if not agent_card:
                    self.log.warning(f"Could not retrieve agent card for {key}, continuing to next agent.")
                    # skip if we couldn't get the agent card
                    continue

                # use custom function metadata builder or default
                build_metadata = (
                    config.build_function_metadata or self._build_function_metadata or self._default_function_metadata
                )

                metadata = build_metadata(agent_card)
                name = metadata.name
                description = metadata.description

                async def message_handler(
                    params: A2AClientMessageParams, config: AgentConfig, agent_card: AgentCard, key: str
                ) -> Any:
                    try:
                        agent_message = params.message

                        # use custom message builder if provided, otherwise use default
                        metadata = BuildMessageForAgentMetadata(card=agent_card, input=agent_message)
                        msg_or_str = (
                            config.build_message_for_agent(metadata)
                            if config.build_message_for_agent
                            else self._build_message_for_agent(metadata)
                            if self._build_message_for_agent
                            else self._default_build_message(agent_card, agent_message)
                        )

                        # handle both message and string returns
                        if isinstance(msg_or_str, str):
                            message = self._create_message(msg_or_str, agent_card)
                        else:
                            message = msg_or_str

                        # get the client info to send the message
                        client_info = self._clients.get(key)
                        if not client_info:
                            raise ValueError(f"Client not found for agent {key}")

                        self.log.debug(f"Calling agent {agent_card.name} with {message.model_dump_json()}")

                        last_message: Optional[Message] = None
                        async for event in client_info.client.send_message(message):
                            # Unwrap tuple from transport implementations
                            if isinstance(event, tuple):
                                event = event[0]
                            if isinstance(event, Message):
                                last_message = event

                        if last_message is not None:
                            self.log.debug(f"Got response from {agent_card.name}")
                            # use custom response builder if provided, otherwise use default
                            metadata = BuildMessageFromAgentMetadata(
                                card=agent_card, response=last_message, original_input=agent_message
                            )
                            return (
                                config.build_message_from_agent_response(metadata)
                                if config.build_message_from_agent_response
                                else self._build_message_from_agent_response(metadata)
                                if self._build_message_from_agent_response
                                else self._default_build_message_from_agent_response(
                                    agent_card, last_message, agent_message
                                )
                            )
                    except Exception as e:
                        self.log.error(e)
                        raise e

                message_handler_with_args = partial(message_handler, config=config, agent_card=agent_card, key=key)

                all_functions.append(
                    ChatFunction(
                        name=name,
                        description=description,
                        parameter_schema=A2AClientMessageParams,
                        handler=message_handler_with_args,
                    )
                )
                self.log.debug(f"Added function in ChatPrompt to call {agent_card.name}")
            except Exception as e:
                self.log.info(f"Failed to build function for agent {key}: {e}")
                # Continue with other agents even if one fails

        functions = functions + all_functions
        return functions

    async def on_build_instructions(self, instructions: Optional[SystemMessage] = None) -> Optional[SystemMessage]:
        # collect agent details for prompt building
        agent_prompt_params: List[AgentPromptParams] = []

        for key, config in self._agent_configs.items():
            try:
                agent_card = await self._get_agent_card(key, config)
                if agent_card:
                    client_info = self._clients.get(key)
                    if client_info:
                        agent_prompt_params.append(
                            AgentPromptParams(
                                card=agent_card,
                                client=client_info.client,
                            )
                        )
            except Exception as e:
                self.log.info(f"Failed to get agent card for {key}: {e}")

        # use custom build_prompt if provided, otherwise use default
        if self._build_prompt:
            metadata = BuildPromptMetadata(
                system_prompt=instructions.content if instructions else None, agent_details=agent_prompt_params
            )
            content = self._build_prompt(metadata)
            return SystemMessage(content=content or "")

        # default prompt building
        if len(agent_prompt_params) == 0:
            return instructions

        agent_details = "".join(
            [
                f"<Agent Details>\n<Name>\n{param.card.name}\n</Name>\n"
                + (f"<Description>\n{param.card.description}\n</Description>\n" if param.card.description else "")
                + "".join(
                    [
                        f'<SKILL name="{skill.name}" description="{skill.description}" />\n'
                        + (f"<EXAMPLES>\n{'\n'.join(skill.examples)}\n</EXAMPLES>\n" if skill.examples else "")
                        for skill in param.card.skills
                    ]
                )
                # could add client specific info here if needed
                + "</Agent Details>\n"
                for param in agent_prompt_params
            ]
        )

        prompt = (
            (instructions.content if instructions else "")
            + "\n\nHere are details about available agents that you can message:\n"
            + agent_details
        )

        return SystemMessage(content=prompt)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .agent_client_info import AgentClientInfo
from .agent_config import AgentConfig
from .plugin import A2AClientPlugin, A2AClientPluginOptions, A2APluginUseParams, FunctionMetadata
from .types import (
    BuildFunctionMetadata,
    BuildMessageForAgent,
    BuildMessageForAgentMetadata,
    BuildMessageFromAgentMetadata,
    BuildMessageFromAgentResponse,
)

__all__ = [
    "AgentClientInfo",
    "AgentConfig",
    "A2AClientPlugin",
    "BuildFunctionMetadata",
    "BuildMessageForAgent",
    "BuildMessageFromAgentResponse",
    "A2AClientPluginOptions",
    "A2APluginUseParams",
    "BuildMessageForAgentMetadata",
    "BuildMessageFromAgentMetadata",
    "FunctionMetadata",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from logging import Logger

from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp


class LoggingMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, logger: Logger) -> None:
        super().__init__(app)
        self.logger = logger

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        log_message = f"A2A Request: {request.method} {request.url}"

        if request.method == "POST":
            body = await request.body()
            log_message += f" Body: {body.decode('utf-8')}"
        self.logger.debug(log_message)
        response = await call_next(request)
        return response



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.


Backward compatibility shim for microsoft.teams.a2a.

DEPRECATED: This import path is deprecated and will be removed in version 2.0.0 GA.
Please update your imports to use 'microsoft_teams.a2a' instead.
"""

import sys
import warnings

warnings.warn(
    "The 'microsoft.teams.a2a' namespace is deprecated and will be removed in "
    "version 2.0.0 GA. Please update your imports to 'microsoft_teams.a2a'.",
    DeprecationWarning,
    stacklevel=2,
)

from microsoft_teams.a2a import *  # noqa: E402, F401, F403
from microsoft_teams.a2a import __all__  # noqa: E402, F401

_new_module = sys.modules["microsoft_teams.a2a"]
sys.modules[__name__] = _new_module



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from typing import Optional

from .types import BuildFunctionMetadata, BuildMessageForAgent, BuildMessageFromAgentResponse


@dataclass
class AgentConfig:
    key: str
    base_url: str
    card_url: str
    build_function_metadata: Optional[BuildFunctionMetadata] = None
    "Optional function to customize the function name and description for each agent card."
    build_message_for_agent: Optional[BuildMessageForAgent] = None
    "Optional function to customize the message format sent to each agent."
    build_message_from_agent_response: Optional[BuildMessageFromAgentResponse] = None
    "Optional function to customize how agent responses are processed into strings."



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Callable, List, Optional, TypedDict, Union

from a2a.client import Client
from a2a.types import AgentCard, Message, Task


@dataclass
class FunctionMetadata:
    name: str
    description: str


@dataclass
class AgentPromptParams:
    card: AgentCard
    client: Client


@dataclass
class BuildPromptMetadata:
    system_prompt: Optional[str] = None
    agent_details: List[AgentPromptParams] = field(default_factory=lambda: [])


@dataclass
class BuildMessageForAgentMetadata:
    card: AgentCard
    input: str
    metadata: Optional[dict[str, Any]] = None


@dataclass
class BuildMessageFromAgentMetadata:
    card: AgentCard
    response: Union[Task, Message]
    original_input: str


BuildFunctionMetadata = Callable[[AgentCard], FunctionMetadata]
BuildPrompt = Callable[[BuildPromptMetadata], Optional[str]]
BuildMessageForAgent = Callable[[BuildMessageForAgentMetadata], Union[Message, str]]
BuildMessageFromAgentResponse = Callable[[BuildMessageFromAgentMetadata], str]


class A2AClientPluginOptions(TypedDict, total=False):
    """
    Options for constructing an A2AClientPlugin using the official SDK.
    """

    build_function_metadata: Optional[BuildFunctionMetadata]
    "Optional function to customize the function name and description for each agent card."
    build_prompt: Optional[BuildPrompt]
    "Optional function to customize the prompt given all agent cards."
    build_message_for_agent: Optional[BuildMessageForAgent]
    "Optional function to customize the message format sent to each agent."
    build_message_from_agent_response: Optional[BuildMessageFromAgentResponse]
    "Optional function to customize how agent responses are processed into strings."
    logger: Optional[Logger]
    "The associated logger"


@dataclass
class InternalA2AClientPluginOptions:
    """Internal dataclass for A2AClientPluginOptions with defaults and non-nullable fields."""

    # Optional fields
    build_function_metadata: Optional[BuildFunctionMetadata] = None
    build_prompt: Optional[BuildPrompt] = None
    build_message_for_agent: Optional[BuildMessageForAgent] = None
    build_message_from_agent_response: Optional[BuildMessageFromAgentResponse] = None
    logger: Optional[Logger] = None

    @classmethod
    def from_typed_dict(cls, options: A2AClientPluginOptions) -> "InternalA2AClientPluginOptions":
        """
        Create InternalA2AClientPluginOptions from A2AClientPluginOptions TypedDict.

        Args:
            options: A2AClientPluginOptions TypedDict

        Returns:
            InternalA2AClientPluginOptions with proper defaults and non-nullable required fields
        """
        kwargs: dict[str, Any] = {k: v for k, v in options.items() if v is not None}
        return cls(**kwargs)


@dataclass
class A2APluginUseParams:
    """
    Parameters for registering an agent with the A2AClientPlugin.
    """

    key: str
    "Unique key to identify this agent"
    base_url: str
    "Base URL to the agent's card"
    card_url: str
    "URL to the agent's card endpoint"
    build_function_metadata: Optional[BuildFunctionMetadata] = None
    "Custom function metadata builder for this specific agent"
    build_message_for_agent: Optional[BuildMessageForAgent] = None
    "Custom message builder for this specific agent"
    build_message_from_agent_response: Optional[BuildMessageFromAgentResponse] = None
    "Custom response processor for this specific agent"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass

from a2a.client import Client
from a2a.types import AgentCard

from .agent_config import AgentConfig


@dataclass(kw_only=True)
class AgentClientInfo(AgentConfig):
    client: Client
    agent_card: AgentCard



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .a2a_plugin_options import A2APluginOptions
from .custom_agent_executor import A2AMessageEvent, A2AMessageEventKey, CustomAgentExecutor, Respond
from .logging_middleware import LoggingMiddleware
from .plugin import A2APlugin

__all__ = [
    "A2APluginOptions",
    "CustomAgentExecutor",
    "LoggingMiddleware",
    "A2APlugin",
    "Respond",
    "A2AMessageEvent",
    "A2AMessageEventKey",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import asyncio
import uuid
from typing import Awaitable, Callable, TypedDict, Union

from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.types import Message, Part, Role, Task, TextPart

type Respond = Callable[[Union[str, Message, Task]], Awaitable[None]]

# This key MUST be used for A2A message events.
A2AMessageEventKey = "a2a:message"


class A2AMessageEvent(TypedDict):
    request_context: RequestContext
    respond: Respond


class CustomAgentExecutor(AgentExecutor):
    def __init__(self, plugin_event: Callable[[str, A2AMessageEvent], Awaitable[None]]):
        super().__init__()
        self.plugin_event = plugin_event

    async def execute(self, context: RequestContext, event_queue: EventQueue) -> None:
        response_future: asyncio.Future[None] = asyncio.Future()

        async def Respond(message: Union[str, Message, Task]) -> None:
            nonlocal response_future
            response_message: Union[Message, Task]

            if isinstance(message, str):
                response_message = Message(
                    kind="message",
                    message_id=str(uuid.uuid4()),
                    role=Role("agent"),
                    parts=[Part(root=TextPart(kind="text", text=message))],
                    # Associate the response with the incoming request's context.
                    context_id=context.context_id,
                )
            else:
                response_message = message

            await event_queue.enqueue_event(response_message)
            response_future.set_result(None)

        ctx: A2AMessageEvent = {
            "request_context": context,
            "respond": Respond,
        }

        if callable(self.plugin_event):
            await self.plugin_event(A2AMessageEventKey, ctx)
            await response_future

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        pass
